"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

require("core-js/modules/es.array.iterator.js");

require("core-js/modules/es.promise.js");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

function _defineProperty2() {
  const data = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

  _defineProperty2 = function () {
    return data;
  };

  return data;
}

function _bluebird() {
  const data = require("bluebird");

  _bluebird = function () {
    return data;
  };

  return data;
}

function _pMapSeries() {
  const data = _interopRequireDefault(require("p-map-series"));

  _pMapSeries = function () {
    return data;
  };

  return data;
}

function _groupArray() {
  const data = _interopRequireDefault(require("group-array"));

  _groupArray = function () {
    return data;
  };

  return data;
}

function _ramda() {
  const data = _interopRequireDefault(require("ramda"));

  _ramda = function () {
    return data;
  };

  return data;
}

function _analytics() {
  const data = require("../../analytics/analytics");

  _analytics = function () {
    return data;
  };

  return data;
}

function _bitId() {
  const data = require("../../bit-id");

  _bitId = function () {
    return data;
  };

  return data;
}

function _generalError() {
  const data = _interopRequireDefault(require("../../error/general-error"));

  _generalError = function () {
    return data;
  };

  return data;
}

function _showDoctorError() {
  const data = _interopRequireDefault(require("../../error/show-doctor-error"));

  _showDoctorError = function () {
    return data;
  };

  return data;
}

function _enrichContextFromGlobal() {
  const data = _interopRequireDefault(require("../../hooks/utils/enrich-context-from-global"));

  _enrichContextFromGlobal = function () {
    return data;
  };

  return data;
}

function _logger() {
  const data = _interopRequireDefault(require("../../logger/logger"));

  _logger = function () {
    return data;
  };

  return data;
}

function _utils() {
  const data = require("../../utils");

  _utils = function () {
    return data;
  };

  return data;
}

function _exceptions() {
  const data = require("../exceptions");

  _exceptions = function () {
    return data;
  };

  return data;
}

function _objects() {
  const data = require("../objects");

  _objects = function () {
    return data;
  };

  return data;
}

function _scopeRemotes() {
  const data = require("../scope-remotes");

  _scopeRemotes = function () {
    return data;
  };

  return data;
}

function _versionDependencies() {
  const data = _interopRequireDefault(require("../version-dependencies"));

  _versionDependencies = function () {
    return data;
  };

  return data;
}

const removeNils = _ramda().default.reject(_ramda().default.isNil);

class ScopeComponentsImporter {
  constructor(scope) {
    (0, _defineProperty2().default)(this, "scope", void 0);
    (0, _defineProperty2().default)(this, "sources", void 0);
    if (!scope) throw new Error('unable to instantiate ScopeComponentsImporter without Scope');
    this.scope = scope;
    this.sources = scope.sources;
  }

  static getInstance(scope) {
    return new ScopeComponentsImporter(scope);
  }
  /**
   * 1. Local objects, fetch from local. (done by this.sources.getMany method)
   * 2. Fetch flattened dependencies (done by toVersionDependencies method). If they're not locally, fetch from remote
   * and save them locally.
   * 3. External objects, fetch from a remote and save locally. (done by this._getExternalMany method).
   */


  async importMany(ids, cache = true, persist = true) {
    _logger().default.debugAndAddBreadCrumb('ScopeComponentsImporter', `importMany: cache ${cache}, ids: {ids}`, {
      ids: ids.toString()
    });

    const idsWithoutNils = removeNils(ids);
    if (_ramda().default.isEmpty(idsWithoutNils)) return Promise.resolve([]);

    const [locals, externals] = _ramda().default.partition(id => id.isLocal(this.scope.name), idsWithoutNils);

    const localDefs = await this.sources.getMany(locals);
    const versionDeps = await (0, _pMapSeries().default)(localDefs, def => {
      if (!def.component) throw new (_exceptions().ComponentNotFound)(def.id.toString());
      return this.componentToVersionDependencies(def.component, def.id);
    });
    const versionDepsWithoutNull = removeNils(versionDeps);

    _logger().default.debugAndAddBreadCrumb('ScopeComponentsImporter', 'importMany: successfully fetched local components and their dependencies. Going to fetch externals');

    const remotes = await (0, _scopeRemotes().getScopeRemotes)(this.scope);
    const externalDeps = await this._getExternalMany(externals, remotes, cache, persist);
    return versionDepsWithoutNull.concat(externalDeps);
  }

  async importManyWithoutDependencies(ids, cache = true) {
    if (!ids.length) return [];

    _logger().default.debug(`importManyWithoutDependencies. Ids: ${ids.join(', ')}, cache: ${cache.toString()}`);

    _analytics().Analytics.addBreadCrumb('importManyWithoutDependencies', `Ids: ${_analytics().Analytics.hashData(ids)}`);

    const idsWithoutNils = removeNils(ids);
    if (_ramda().default.isEmpty(idsWithoutNils)) return Promise.resolve([]);
    const [externals, locals] = (0, _utils().splitBy)(idsWithoutNils, id => id.isLocal(this.scope.name));
    const localDefs = await this.sources.getMany(locals);
    const componentVersionArr = await Promise.all(localDefs.map(def => {
      if (!def.component) throw new (_exceptions().ComponentNotFound)(def.id.toString()); // $FlowFixMe it must have a version
      // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!

      return def.component.toComponentVersion(def.id.version);
    }));
    const remotes = await (0, _scopeRemotes().getScopeRemotes)(this.scope);
    const externalDeps = await this._getExternalManyWithoutDependencies(externals, remotes, cache);
    return componentVersionArr.concat(externalDeps);
  }
  /**
   * todo: improve performance by finding all versions needed and fetching them in one request from the server
   * currently it goes to the server twice. First, it asks for the last version of each id, and then it goes again to
   * ask for the older versions.
   */


  async importManyWithAllVersions(ids, cache = true, allDepsVersions = false // by default, only dependencies of the latest version are imported
  ) {
    _logger().default.debug(`scope.getManyWithAllVersions, Ids: ${ids.join(', ')}`);

    _analytics().Analytics.addBreadCrumb('getManyWithAllVersions', `scope.getManyWithAllVersions, Ids: ${_analytics().Analytics.hashData(ids)}`);

    const idsWithoutNils = removeNils(ids);
    if (_ramda().default.isEmpty(idsWithoutNils)) return Promise.resolve([]);
    const versionDependenciesArr = await this.importMany(idsWithoutNils, cache);
    const allIdsWithAllVersions = new (_bitId().BitIds)();
    versionDependenciesArr.forEach(versionDependencies => {
      const versions = versionDependencies.component.component.listVersions();
      const versionId = versionDependencies.component.id;
      const idsWithAllVersions = versions.map(version => {
        if (version === versionDependencies.component.version) return null; // imported already

        return versionId.changeVersion(version);
      });
      allIdsWithAllVersions.push(...removeNils(idsWithAllVersions));
      const head = versionDependencies.component.component.getHead();

      if (head) {
        allIdsWithAllVersions.push(versionId.changeVersion(head.toString()));
      }
    });

    if (allDepsVersions) {
      const verDepsOfOlderVersions = await this.importMany(allIdsWithAllVersions, cache);
      versionDependenciesArr.push(...verDepsOfOlderVersions);
      const allFlattenDepsIds = versionDependenciesArr.map(v => v.allDependencies.map(d => d.id));

      const dependenciesOnly = _ramda().default.flatten(allFlattenDepsIds).filter(id => !ids.hasWithoutVersion(id));

      const verDepsOfAllFlattenDeps = await this.importManyWithAllVersions(_bitId().BitIds.uniqFromArray(dependenciesOnly));
      versionDependenciesArr.push(...verDepsOfAllFlattenDeps);
    } else {
      await this.importManyWithoutDependencies(allIdsWithAllVersions);
    }

    return versionDependenciesArr;
  }

  async importFromLanes(remoteLaneIds) {
    const lanes = await this.importLanes(remoteLaneIds);
    const ids = lanes.map(lane => lane.toBitIds());

    const bitIds = _bitId().BitIds.uniqFromArray(_ramda().default.flatten(ids));

    await this.importManyWithAllVersions(bitIds, false);
    return lanes;
  }

  async importLanes(remoteLaneIds) {
    const remotes = await (0, _scopeRemotes().getScopeRemotes)(this.scope);
    const objectList = await remotes.fetch(groupByScopeName(remoteLaneIds), this.scope, {
      type: 'lane'
    });
    const bitObjects = await objectList.toBitObjects();
    const lanes = bitObjects.getLanes();
    await Promise.all(lanes.map(lane => this.scope.objects.remoteLanes.syncWithLaneObject(lane.scope, lane)));
    return lanes;
  }

  async componentToVersionDependencies(component, id, throwForNoVersion = false) {
    if (!throwForNoVersion) {
      if (component.isEmpty() && !id.hasVersion() && !component.laneHeadLocal) {
        // this happens for example when importing a remote lane and then running "bit fetch --components"
        // the head is empty because it exists on the lane only, it was never tagged and
        // laneHeadLocal was never set as it originated from the scope, not the consumer.
        return null;
      }
    } // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!


    const versionComp = component.toComponentVersion(id.version); // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!

    const source = id.scope;
    const version = await versionComp.getVersion(this.scope.objects);

    if (!version) {
      _logger().default.debug(`toVersionDependencies, component ${component.id().toString()}, version ${versionComp.version} not found, going to fetch from a remote`);

      if (component.scope === this.scope.name) {
        // it should have been fetched locally, since it wasn't found, this is an error
        throw new (_showDoctorError().default)(`Version ${versionComp.version} of ${component.id().toString()} was not found in scope ${this.scope.name}`);
      }

      return (0, _scopeRemotes().getScopeRemotes)(this.scope).then(remotes => {
        return this._getExternal({
          id,
          remotes,
          localFetch: false
        });
      });
    }

    _logger().default.debug(`toVersionDependencies, component ${component.id().toString()}, version ${versionComp.version} found, going to collect its dependencies`);

    const dependencies = await this.importManyWithoutDependencies(version.flattenedDependencies);
    return new (_versionDependencies().default)(versionComp, dependencies, [], [], source);
  }

  async componentsToComponentsObjects(components, clientVersion, collectParents, collectArtifacts) {
    const allObject = await (0, _pMapSeries().default)(components, component => component.toObjects(this.scope.objects, clientVersion, collectParents, collectArtifacts));
    return _ramda().default.flatten(allObject);
  }
  /**
   * get ConsumerComponent by bitId. if the component was not found locally, import it from a remote scope
   */


  loadRemoteComponent(id) {
    return this._getComponentVersion(id).then(component => {
      if (!component) throw new (_exceptions().ComponentNotFound)(id.toString()); // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!

      return component.toConsumer(this.scope.objects);
    });
  }

  loadComponent(id, localOnly = true) {
    _logger().default.debugAndAddBreadCrumb('ScopeComponentsImporter', 'loadComponent {id}', {
      id: id.toString()
    });

    if (localOnly && !id.isLocal(this.scope.name)) {
      throw new (_generalError().default)('cannot load a component from remote scope, please import first');
    }

    return this.loadRemoteComponent(id);
  }
  /**
   * recursive function.
   * if found locally, use them. Otherwise, fetch from remote and then, save into the model.
   */


  async _getExternalMany(ids, remotes, localFetch = true, persist = true, context = {}) {
    if (!ids.length) return Promise.resolve([]);

    _logger().default.debugAndAddBreadCrumb('scope.getExternalMan', `planning on fetching from ${localFetch ? 'local' : 'remote'} scope. Ids: {ids}`, {
      ids: ids.join(', ')
    });

    (0, _enrichContextFromGlobal().default)(Object.assign(context, {
      requestedBitIds: ids.map(id => id.toString())
    }));
    const componentDefs = await this.sources.getMany(ids);
    const left = componentDefs.filter(componentDef => {
      if (!localFetch) return true;
      if (!componentDef.component) return true;
      return false;
    });

    if (left.length === 0) {
      _logger().default.debugAndAddBreadCrumb('scope.getExternalMany', 'no more ids left, all found locally, exiting the method');

      const versionDeps = await (0, _pMapSeries().default)(componentDefs, compDef => this.componentToVersionDependencies(compDef.component, compDef.id));
      return removeNils(versionDeps);
    }

    _logger().default.debugAndAddBreadCrumb('scope.getExternalMany', `${left.length} left. Fetching them from a remote`);

    const objectList = await remotes.fetch(groupByScopeName(left.map(defLeft => defLeft.id)), this.scope, undefined, context);

    _logger().default.debugAndAddBreadCrumb('scope.getExternalMany', 'writing them to the model');

    const nonLaneIds = await this.scope.writeManyComponentsToModel(objectList, persist, ids);
    return this._getExternalMany(nonLaneIds, remotes);
  }
  /**
   * If the component is not in the local scope, fetch it from a remote and save into the local
   * scope. (objects directory).
   */


  async _getExternal({
    id,
    remotes,
    localFetch = true,
    context = {}
  }) {
    (0, _enrichContextFromGlobal().default)(context);
    const component = await this.sources.get(id);

    if (component && localFetch) {
      const versionDeps = await this.componentToVersionDependencies(component, id, true);
      return versionDeps;
    }

    const objectList = await remotes.fetch(groupByScopeName([id]), this.scope, undefined, context);
    await this.scope.writeManyComponentsToModel(objectList, true, [id]);
    return this._getExternal({
      id,
      remotes,
      localFetch: true
    });
  }

  async _getExternalWithoutDependencies({
    id,
    remotes,
    localFetch = true,
    context = {}
  }) {
    const component = await this.sources.get(id);

    if (component && localFetch) {
      return component.toComponentVersion(id.version);
    }

    const objectList = await remotes.fetch(groupByScopeName([id]), this.scope, {
      withoutDependencies: true
    }, context);
    await this.scope.writeManyComponentsToModel(objectList, true, [id]);
    const versionDependencies = await this._getExternal({
      id,
      remotes,
      localFetch: true
    });
    return versionDependencies.component;
  }

  _getExternalManyWithoutDependencies(ids, remotes, localFetch = false, context = {}) {
    if (!ids.length) return Promise.resolve([]);

    _logger().default.debugAndAddBreadCrumb('_getExternalManyWithoutDependencies', `ids: {ids}, localFetch: ${localFetch.toString()}`, {
      ids: ids.join(', ')
    });

    (0, _enrichContextFromGlobal().default)(Object.assign(context, {
      requestedBitIds: ids.map(id => id.toString())
    }));
    return this.sources.getMany(ids).then(defs => {
      const left = defs.filter(def => {
        if (!localFetch) return true;
        if (!def.component) return true;
        return false;
      });

      if (left.length === 0) {
        _logger().default.debugAndAddBreadCrumb('_getExternalManyWithoutDependencies', 'no more ids left, all found locally, exiting the method'); // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!


        return Promise.all(defs.map(def => def.component.toComponentVersion(def.id.version)));
      }

      _logger().default.debugAndAddBreadCrumb('_getExternalManyWithoutDependencies', `${left.length} left. Fetching them from a remote`);

      return remotes.fetch(groupByScopeName(left.map(def => def.id)), this.scope, {
        withoutDependencies: true
      }, context).then(objectList => {
        return this.scope.writeManyComponentsToModel(objectList, undefined, ids);
      }).then(nonLaneIds => this._getExternalManyWithoutDependencies(nonLaneIds, remotes, true));
    });
  }

  async _getComponentVersion(id) {
    if (!id.isLocal(this.scope.name)) {
      const remotes = await (0, _scopeRemotes().getScopeRemotes)(this.scope);
      return this._getExternalWithoutDependencies({
        id,
        remotes,
        localFetch: true
      });
    }

    return this.sources.get(id).then(component => {
      if (!component) throw new (_exceptions().ComponentNotFound)(id.toString()); // $FlowFixMe version is set
      // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!

      return component.toComponentVersion(id.version);
    });
  }
  /**
   * currently used for import artifacts, but can be used to import any arbitrary array of hashes.
   * it takes care to remove any duplications and check whether the object exists locally before
   * going to the remote
   */


  async importManyObjects(groupedHashes) {
    const groupedHashedMissing = {};
    await Promise.all(Object.keys(groupedHashes).map(async scopeName => {
      const uniqueHashes = _ramda().default.uniq(groupedHashes[scopeName]);

      const missing = await (0, _bluebird().filter)(uniqueHashes, async hash => !(await this.scope.objects.has(new (_objects().Ref)(hash))));

      if (missing.length) {
        groupedHashedMissing[scopeName] = missing;
      }
    }));
    if (_ramda().default.isEmpty(groupedHashedMissing)) return;
    const remotes = await (0, _scopeRemotes().getScopeRemotes)(this.scope);
    const objectList = await remotes.fetch(groupedHashedMissing, this.scope, {
      type: 'object'
    });
    const bitObjectsList = await objectList.toBitObjects();
    this.scope.objects.addMany(bitObjectsList.getAll());
    await this.scope.objects.persist();
  }

}

exports.default = ScopeComponentsImporter;

function groupByScopeName(ids) {
  const grouped = (0, _groupArray().default)(ids, 'scope');
  Object.keys(grouped).forEach(scopeName => {
    grouped[scopeName] = grouped[scopeName].map(id => id.toString());
  });
  return grouped;
}