"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

require("core-js/modules/es.array.iterator.js");

require("core-js/modules/es.promise.js");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

function _defineProperty2() {
  const data = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

  _defineProperty2 = function () {
    return data;
  };

  return data;
}

function _ramda() {
  const data = _interopRequireDefault(require("ramda"));

  _ramda = function () {
    return data;
  };

  return data;
}

function _semver() {
  const data = _interopRequireDefault(require("semver"));

  _semver = function () {
    return data;
  };

  return data;
}

function _bitId() {
  const data = require("../bit-id");

  _bitId = function () {
    return data;
  };

  return data;
}

function _customError() {
  const data = _interopRequireDefault(require("../error/custom-error"));

  _customError = function () {
    return data;
  };

  return data;
}

function _showDoctorError() {
  const data = _interopRequireDefault(require("../error/show-doctor-error"));

  _showDoctorError = function () {
    return data;
  };

  return data;
}

function _logger() {
  const data = _interopRequireDefault(require("../logger/logger"));

  _logger = function () {
    return data;
  };

  return data;
}

function _traverseVersions() {
  const data = require("./component-ops/traverse-versions");

  _traverseVersions = function () {
    return data;
  };

  return data;
}

function _exceptions() {
  const data = require("./exceptions");

  _exceptions = function () {
    return data;
  };

  return data;
}

class ComponentVersion {
  constructor(component, version) {
    (0, _defineProperty2().default)(this, "component", void 0);
    (0, _defineProperty2().default)(this, "version", void 0);

    if (!version) {
      throw new TypeError(`ComponentVersion expects "version" to be defined (failed for ${component.id()})`);
    }

    this.component = component;
    this.version = version;
    Object.freeze(this);
  }

  getVersion(repository) {
    return this.component.loadVersion(this.version, repository);
  }

  flattenedDependencies(repository) {
    return this.getVersion(repository).then(version => version.flattenedDependencies);
  }

  toId() {
    return new (_bitId().BitId)({
      scope: this.component.scope,
      name: this.component.name,
      version: this.version
    });
  } // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!


  get id() {
    return this.toId();
  }

  toConsumer(repo, manipulateDirData) {
    // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!
    return this.component.toConsumerComponent(this.version, this.component.scope, repo, manipulateDirData);
  }

  async toObjects(repo, clientVersion, collectParents, collectArtifacts) {
    const version = await this.getVersion(repo);
    if (!version) throw new (_showDoctorError().default)(`failed loading version ${this.version} of ${this.component.id()}`); // @todo: remove this customError once upgrading to v15, because when the server has v15
    // and the client has < 15, the client will get anyway an error to upgrade the version

    if (clientVersion && version.overrides && !_ramda().default.isEmpty(version.overrides) && _semver().default.lt(clientVersion, '14.1.0')) {
      throw new (_customError().default)(`Your components were created with a newer version and use the "overrides" feature.
Please upgrade your bit client to version >= v14.1.0`);
    }

    const collectVersionObjects = async ver => {
      const versionRefs = ver.refsWithOptions(collectParents, collectArtifacts);
      const versionObjects = await ver.collectManyObjects(repo, versionRefs);
      const versionData = {
        ref: ver.hash(),
        buffer: await ver.asRaw(repo)
      };
      return [...versionObjects, versionData];
    };

    try {
      const componentData = {
        ref: this.component.hash(),
        buffer: await this.component.asRaw(repo)
      };
      const parentsObjects = [];

      if (collectParents) {
        const allParentsHashes = await (0, _traverseVersions().getAllVersionHashes)(this.component, repo, true, version.hash());
        const missingParentsHashes = allParentsHashes.filter(h => !h.isEqual(version.hash()));
        missingParentsHashes.map(async parentHash => {
          const parentVersion = await parentHash.load(repo);
          const parentsObj = await collectVersionObjects(parentVersion);
          parentsObjects.push(...parentsObj);
        });
      }

      const versionObjects = await collectVersionObjects(version);
      const scopeMeta = await repo.getScopeMetaObject();
      return [componentData, ...versionObjects, ...parentsObjects, scopeMeta];
    } catch (err) {
      _logger().default.error(`component-version.toObjects ${this.id.toString()} got an error`, err); // @ts-ignore


      const originalVersionHash = this.component.getRef(this.version).toString();
      const currentVersionHash = version.hash().toString();

      if (originalVersionHash !== currentVersionHash) {
        throw new (_exceptions().HashMismatch)(this.component.id(), this.version, originalVersionHash, currentVersionHash);
      }

      throw err;
    }
  }

}

exports.default = ComponentVersion;